// cforth.h Header file of the CForth  project ********************************
//
// Author and (c)2020-24 Wolfgang Schemmert. The copyright claim is restricted
// to this file and the depending files "main.c", "main.h", "CForth.c"
// and "globaldefines.h"
// Other depending material is published, copyrighted and licensed
// by STMicroelectronics and Segger GmbH
//
// Contact: <www.midi-and-more.de>
// A detailled hardware construction and operation manual
// is provided at this website
//
// ****************************************************************************
//      This program is free software: you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation, either version 3 of the License, or
//      (at your option) any later version.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//      Licence: GPL-3
//
// ****************************************************************************

//from main.h
extern uint8_t *gUSART_RXHEAD, *gUSART_RXTAIL;
extern uint32_t gSYSCOUNT, gTIMEOUT;
extern uint32_t gProject;
extern const uint8_t gTypeStr [];
extern const uint8_t gClockStr [];

extern void sendModel(void);
extern void popFlashPage23(void);
extern uint32_t pushFlashPage23(void);
extern void popFlashProject();
extern uint32_t pushFlashProject(void);
extern void msDelay(uint32_t);
extern void wait5us(void);
extern void servSystem(void);
extern void initUsart2(void);
extern void sendTB(uint8_t);
extern void sendString(const uint8_t *);
extern void sendNibble(uint8_t);
extern void sendX8(uint32_t);
extern void sendX32(uint32_t);
extern void sendNX32(uint32_t);
extern void send0X32(uint32_t);
extern void sendLS32(uint32_t);
extern void sendS32(uint32_t);
extern void sendNS32(uint32_t);
extern uint32_t pullTermRxBuf(void);

// #########################################################

#define IDX_CONTI		0x72
#define IDX_BREAK		0x73
#define IDX_ZBREAK		0x74
#define IDX_SEMIS	    0x77
#define IDX_DOSTR	    0x78
#define IDX_DOLIT		0x79
#define IDX_DOVALUE		0x7A
#define IDX_DOREV		0x7B
#define IDX_DOIF        0x7C
#define IDX_DOCASE      0x7D
#define IDX_DOLTCAS     0x7E
#define IDX_DOGTCAS     0x7F
#define IDX_DOELSE      0x80
#define IDX_DODO	    0x81
#define IDX_DOUNTIL		0x82
#define IDX_DOWHILE		0x83
#define IDX_DOAGAIN     0x84
#define IDX_DOFOR		0x85
#define IDX_DOLOOP		0x86
#define IDX_RPDEC		0x87
#define IDX_RP4DEC		0x88

//arithmetic ops
void op_ERR(void);
void op_DROP(void);
void op_DUP(void);
void op_SWAP(void);
void op_OVER(void);
void op_ROT(void);
void op_DNROT(void);
void op_PICK(void);
void op_ROLL(void);
void op_INJECT(void);
void op_PLUS(void);
void op_INCT(void);
void op_MINUS(void);
void op_DECT(void);
void op_ABS(void);
void op_SHIFT(void);
void op_AND(void);
void op_OR(void);
void op_XOR(void);
void op_NOT(void);
void op_NEG(void);
void op_MULT(void);
void op_DIV(void);
void op_MULTDIV(void);
void op_MOD(void);
void op_DIVMOD(void);
void op_BSET(void);
void op_BCLR(void);
void op_BTST(void);
void op_RANDOM(void);

//comparison ops
void op_EQUAL(void);
void op_UNEQU(void);
void op_ZEQ(void);
void op_GRTR(void);
void op_GRTREQ(void);
void op_LESS(void);
void op_ULESS(void);
void op_LESSEQ(void);

//memory ops
void op_VARCON(void);
void op_REV(void);
void op_WRITE(void);
void op_READ(void);
void op_ARRAY(void);

//terminal ops
void op_KEYQ(void);
void op_KEY(void);
void op_EMIT(void);
void op_PRINT(void);
void op_PRINTHEX(void);
void op_PRINTDEC(void);
void op_PRINTSTR(void);
void op_DECIM(void);
void op_HEX(void);

//time ops
void op_XNOP(void);
void op_MS(void);
void op_DECIS(void);
void op_TIX(void);
void op_TIME(void);
void op_FREQ12();
void op_PWM1();
void op_PWM2();
void op_FREQ3();
void op_PWM3();

//peripheral ops
void op_PINOH(void);
void op_PINOL(void);
void op_DAO4(void);
void op_DAO5(void);
void op_PINIPU(void);
void op_PINIPD(void);
void op_PINIZ(void);
void op_READDIG(void);
void op_READADC(void);
void op_INITSPI(void);
void op_SPI(void);
void op_INITI2C(void);
void op_I2CTX1(void);
void op_I2CTXN(void);
void op_I2CRX1(void);
void op_I2CRXN(void);
void op_I2CADRXN(void);
void op_MOT1(void);
void op_MOT2(void);
void op_MOT3(void);
void op_REGWRITE(void);
void op_REGREAD(void);
void op_REGBSET(void);
void op_REGBCLR(void);
void op_REGBTST(void);

//structuring ops
void op_IF(void);
void op_CASE(void);
void op_LTCASE(void);
void op_GTCASE(void);
void op_ELSE(void);
void op_THEN(void);
void op_DO(void);
void op_UNTIL(void);
void op_WHILE(void);
void op_AGAIN(void);
void op_FOR(void);
void op_IDX(void);
void op_LOOP(void);
void op_CONTI(void);
void op_BREAK(void);
void op_ZBREAK(void);
void op_RETURN(void);

//compiler ops
void op_PRINTSTACK(void);
void op_FLUSH(void);
void op_OPS(void);
void op_USER(void);
void op_SEE(void);
void op_MEM(void);
void op_STEP(void);
void op_RUN(void);
void op_NODBG(void);
void op_DEBUG(void);
void op_FORGET(void);
void op_ABORT(void);
void op_REPLACE(void);
void op_RESTORE(void);
void op_COLON(void);
void op_SEMIS(void);

//runtime ops
void op_DOSTRING(void);
void op_DOLIT(void);
void op_DOVALUE(void);
void op_DOREV(void);
void op_DOIF(void);
void op_DOCASE(void);
void op_DOLTCAS(void);
void op_DOGTCAS(void);
void op_DOELSE(void);
void op_DODO(void);
void op_DOUNTIL(void);
void op_DOWHILE(void);
void op_DOAGAIN(void);
void op_DOFOR(void);
void op_DOLOOP(void);
void op_DOCONTI(void);
void op_DOBREAK(void);
void op_DOZBREAK(void);
void op_RPDEC(void);
void op_RP4DEC(void);

//system ops
void op_AUTO(void);
void op_QUEST(void);
void op_SAVEFLASH(void);
void op_LOADFLASH(void);

uint32_t gfMSCOUNT, gfTIXCOUNT;                   //1ms countdown in "SysTick_Handler()"
uint8_t gfTib[164], gfTokenstr[80];
uint32_t gfTibMax;
uint32_t gfTokenStart, gfTibPos;
int32_t gfTokenNum;     //simplifies parameter passing in "interpretNumber()"
uint32_t gfErrors=0;
uint32_t gfNumBase=10, gfTempNumBase, gfAuto;
uint32_t gfCompiState=0;//0=interpreting, 1=compiling
uint32_t gfExecUser=0;      //for query(): 1=User token is executing, 0=else
uint32_t gfBackProcData, gfBackNumValid;//primitive 1 level data stack
uint32_t gfTermCheck;	//message from checkTerminal() to KEY?
uint8_t gfTermChar;	//last terminal char while (gfBackProcess && gfExecUser)
uint32_t gfComment=0;   //(!=0):ignore following TIB input, 0:evaluate
                        //gets set to 0 by every CRR, ie valid only per line
uint32_t gfCodeIndex=0, gfCodeStartIndex=0;	    // code Startvalue
uint32_t gfNextNameCompi=0; //gfUserHeader[] next entry index
uint32_t gfParsed;
uint32_t gfBackProcess=1;	
							//init here as precaution,op_ABORT()->final init
uint32_t gfTermInput;		//message from terminal iput to DEBUG();
uint32_t *gfIPold;
uint32_t gfNestLevel, gfBackNestLevel;
uint32_t gfIpVal;
uint32_t gfRandom;
uint32_t gfSpiBits, gfSpiMode, gfSpiClock;
int32_t  gfPWM1, gfPWM2, gfPWM3;
uint32_t gfFreq12, gfTime, gfPuls, gfFreq3;
int32_t gfStepSpeed1, gfStepPhase1, gfSTEPCOUNT1;
int32_t gfStepSpeed2, gfStepPhase2, gfSTEPCOUNT2;
int32_t gfStepSpeed3, gfStepPhase3, gfSTEPCOUNT3;

uint32_t gfDataStack[MAXSTACK];
uint32_t gfReturnStack[MAXSTACK];
uint32_t gfTCode[MAXCODE];
		//=3072words(0xC00)=12288(0x3000)bytes ->total project<=16kBytes

uint32_t *gfSP;				//Datastack Pointer
uint32_t *gfRP;		        //Returnstack Pointer
uint32_t *gfIP;		        //Instruction Pointer

const uint8_t gfKernelNames [KERNELSIZE][8] = {
	{"ERROR"},						//index 0 = dummy for compile error detection
	{"DROP"},{"DUP"},{"SWAP"},{"OVER"},					//index 1..4
	{"ROT"},{"-ROT"},{"PICK"},{"ROLL"},{"INJECT"},		//5...9
	{"+"},{"1+"},{"-"},{"1-"},{"ABS"},					//10...14
	{"SHIFT"},{"AND"},{"OR"},{"XOR"},{"NOT"},			//15...19 =0x0F...0x13
	{"+/-"},{"*"},{"/"},{"*/"},{"MOD"},					//20. .24 =0x14...0x18
	{"/MOD"},{"BSET"},{"BCLR"},{"BTST"},{"RANDOM"},		//25...29 =0x19...0x1D
	{"=="},{"!="},{">"},{">="},{"<"},					//30...34 =0x1E...0x22
	{"U<"},{"<="},{"0="},{"VARCON"},{"REV"},			//35...39 =0x23...0x27
	{"W"},{"R"},{"ARRAY"},{"KEY?"},{"KEY"},				//40...44 =0x28...0x2C
	{"EMIT"},{"."},{".H"},{".D"},{".\""},				//45...49 =0x2D...0x31
	{"DZ"},{"HX"},{"XNOP"},{"MS"},{"DS"},				//50...54 =0x32...0x36
	{"TIX"},{"TIME"},{"FREQ12"},{"PWM1"},{"PWM2"},		//55...59 =0x37...0x3B
	{"FREQ3"},{"PWM3"},{"OH"},{"OL"},{"DAO4"},			//60...64 =0x3C...0x40
	{"DAO5"},{"IPU"},{"IPD"},{"IZ"},{"RDIN"},			//65...69 =0x41...0x45
	{"RADC"},{"INI-SPI"},{"SPI"},{"INI-I2C"},{"ITX1"},	//70...74 =0x46...0x4A
	{"ITXN"},{"IRX1"},{"IRXN"},{"IRXADN"},{"MOT1"},		//75...79 =0x4B...0x4F
	{"MOT2"},{"MOT3"},{"REGW"},{"REGR"},{"REGBSET"},	//80...84.=0x50...0x54
	{"REGBCLR"},{"REGBTST"},{".S"},{"FS"},{"OPS"},		//85...89 =0x55...0x59
	{"USER"},{"SEE"},{"MEM"},{"ST"},{"RUN"},			//90...94 =0x5A...0x5E
	{"NO-DBG"},{"DEBUG"},{"FORGET"},{"ABORT"},{"REPLACE"},//95...99=0x5F..0x63
	{"RESTORE"},{"IF"},{"CASE"},{"<CASE"},{">CASE"},	//100..104=0x64...0x68
	{"ELSE"},{"THEN"},{"DO"},{"UNTIL"},{"WHILE"},		//105...109=0x69..0x6D
	{"AGAIN"},{"FOR"},{"LOOP"},{"I"},{"CONTI"},			//110...114=0x6E..0x72
	{"BREAK"},{"ZBREAK"},{"RETURN"},{":"},{";"},		//115...119=0x73..0x77
	{"doSTR"},{"doLIT"},{"doVAL"},{"doREV"},{"doIF"},	//120...124=0x78..0x7C
	{"doCASE"},{"do<CASE"},{"do>CASE"},{"doELSE"},{"doDO"},//125-129=0x7D-0x81
	{"doUNTIL"},{"doWHILE"},{"doAGAIN"},{"doFOR"},{"doLOOP"},//130-134=0x82-0x86
	{"RP-1"},{"RP-4"},{"AUTOEXE"},{"?"},{"SAVE"},		//135..139=.0x87..0x8B
	{"LOAD"}											//140=.0x8C
};

void (* gfKernelOps [KERNELSIZE]) () = {
op_ERR,op_DROP, op_DUP, op_SWAP, op_OVER,				//index 0..4
op_ROT, op_DNROT, op_PICK, op_ROLL, op_INJECT,			//5...9
op_PLUS, op_INCT, op_MINUS, op_DECT, op_ABS,			//10...14
op_SHIFT, op_AND, op_OR, op_XOR, op_NOT,				//15...19 =0x0F...0x13
op_NEG,op_MULT, op_DIV, op_MULTDIV, op_MOD, 			//20...24 =0x14...0x18
op_DIVMOD, op_BSET, op_BCLR, op_BTST, op_RANDOM, 		//25...29 =0x19...0x1D
op_EQUAL, op_UNEQU, op_GRTR, op_GRTREQ, op_LESS, 		//30...34 =0x1E...0x22
op_ULESS, op_LESSEQ, op_ZEQ, op_VARCON, op_REV, 		//35...39 =0x23...0x27
op_WRITE, op_READ, op_ARRAY, op_KEYQ, op_KEY, 			//40...44 =0x28...0x2C
op_EMIT, op_PRINT, op_PRINTHEX, op_PRINTDEC,op_PRINTSTR,//45...49= 0x2D...0x31
op_DECIM, op_HEX, op_XNOP, op_MS, op_DECIS, 			//50...54 =0x32...0x36
op_TIX, op_TIME, op_FREQ12, op_PWM1, op_PWM2, 			//55...59 =0x37...0x3B
op_FREQ3, op_PWM3, op_PINOH, op_PINOL, op_DAO4, 		//60...64 =0x3C...0x40
op_DAO5, op_PINIPU, op_PINIPD, op_PINIZ, op_READDIG, 	//65...69 =0x41...0x45
op_READADC, op_INITSPI, op_SPI, op_INITI2C, op_I2CTX1,	//70...74 =0x46...0x4A
op_I2CTXN, op_I2CRX1, op_I2CRXN, op_I2CADRXN, op_MOT1, 	//75...79 =0x4B...0x4F
op_MOT2, op_MOT3, op_REGWRITE, op_REGREAD, op_REGBSET,	//80.l.84l=0x50l..0x54
op_REGBCLR,op_REGBTST, op_PRINTSTACK, op_FLUSH, op_OPS, //85...89 =0x55...0x59
op_USER, op_SEE, op_MEM, op_STEP, op_RUN,				//90...94 =0x5A...0x5E
op_NODBG, op_DEBUG, op_FORGET, op_ABORT, op_REPLACE, 	//95...99 =0x5F...0x63
op_RESTORE, op_IF, op_CASE, op_LTCASE, op_GTCASE, 		//100...104=0x64..0x68
op_ELSE, op_THEN, op_DO, op_UNTIL, op_WHILE, 			//105...109=0x69..0x6D
op_AGAIN, op_FOR, op_LOOP, op_IDX, op_CONTI,			//110...114=0x6E..0x72
op_BREAK, op_ZBREAK, op_RETURN, op_COLON, op_SEMIS, 	//115...119=0x73..0x77
op_DOSTRING, op_DOLIT, op_DOVALUE, op_DOREV, op_DOIF, 	//120...124=0x78..0x7C
op_DOCASE, op_DOLTCAS, op_DOGTCAS, op_DOELSE, op_DODO,	//125...129=0x7D..0x81
op_DOUNTIL, op_DOWHILE, op_DOAGAIN, op_DOFOR, op_DOLOOP,//130...134=0x82..0x86
op_RPDEC, op_RP4DEC, op_AUTO,op_QUEST, op_SAVEFLASH, 	//135...139=0x87..0x8B
op_LOADFLASH											//140=0x8C
};

const uint32_t gfKernelAttributes [KERNELSIZE] = {
	0,0,0,0,0,0,0,0,0,0,				//index 0...9
	0,0,0,0,0,0,0,0,0,0,				//10...19 =0x0A...0x13
	0,0,0,0,0,0,0,0,0,0,				//20...29 =0x14...0x1D
	0,0,0,0,0,							//30...34 =0x1E...0x22
	0,0,0,A_IME,A_IB,					//35...39 =0x23...0x27
	0,0,0,0,0,							//40...44 =0x28...0x2C
	0,0,0,0,A_IME,						//45...49 =0x2D...0x31
	A_IB,A_IB,0,0,0,					//50...54 =0x32...0x36
	0,0,A_BAK,A_BAK,A_BAK,				//55...59 =0x37...0x3B
	A_BAK,A_BAK,0,0,0,					//60...64 =0x3C...0x40
	0,0,0,0,0,							//65...69 =0x41...0x45
	0,0,0,0,0,							//70...74 =0x46...0x4A
	0,0,0,0,A_BAK,						//75...79 =0x4B...0x4F
	A_BAK,A_BAK,0,0,0,					//80...84 =0x50...0x54
	0,0,A_BAK,0,A_IB,					//85...89 =0x55...0x59
	A_IB,A_IB,A_BAK,A_BAK,A_BAK,		//90...94 =0x5A...0x5E
	A_BAK,0,A_IME,A_IME,A_IME,			//95...99 =0x5F...0x63
	A_IME,A_CI,A_CI,A_CI,A_CI,			//100..104 =0x64..0x68
	A_CI,A_CI,A_CI,A_CI,A_CI,			//105..109 =0x69..0x6D
	A_CI,A_CI,A_CI,A_CPO,A_CPO,			//110..114 =0x6E..0x72
	A_CPO,A_CPO,A_CPO,A_IME,A_IME,		//115..119 =0x73..0x77
	A_RUN,A_RUN,A_RUN,A_RUN,A_RUN,		//120..124 =0x76..0x7C
	A_RUN,A_RUN,A_RUN,A_RUN,A_RUN,		//125..129 =0x7D..0x81
	A_RUN,A_RUN,A_RUN,A_RUN,A_RUN,		//130...134=0x82..0x86
	A_CPO,A_CPO,A_IME,A_IB,A_IME,		//135...139=0x87..0x8B
	A_IME								//140=0x8C
};

struct USERHEADER
{
uint8_t name[16];
uint32_t toCompile;
uint32_t attributes;
} gfUserHeader[MAXUSER];
//6 words = 24 bytes/struct * 168 entries = 1008 words = 4032 bytes)

void quit(void);
void query(void);
uint32_t subQuery(void);
void eval(void);
void backProcEval(void);
void subPrintString(uint32_t);
void scanNext(void);
uint32_t interpretNumber(void);     //if success, result comes in "gfTokenNum"
uint32_t string2Int(uint8_t*, uint32_t);
uint32_t checkDigit(uint8_t, uint32_t);
uint32_t findKernelToken(void);
uint32_t findUserToken(void);
void interpretKernelToken(void);
uint32_t interpretUserToken(void);
uint32_t compileUserSymbol(uint32_t);
void compileTCode(uint32_t);
void executeTCode(void);
void intoUserCode(void);
void showPreStep(void);
void execNext(void);
void showPostStep(void);
void emptyForth(void);
//void subPrintStack(void);
uint32_t checkTerminal(void);
uint32_t scanString(void);
void subFREQ12();
void subPWM1();
void subPWM2();
void subFREQ3(void);
void subPWM3(void);
uint32_t checkValidIO(void);
void checkSpecialIO(uint32_t);
void initADC(void);
void disableADC(void);
void subInitSpi(void);
void subMOT1(void);
void subMOT2(void);
void subMOT3(void);
void token2String(uint32_t);
void index2String(uint32_t);
uint32_t getToken(void);
uint32_t conFirm(void);
